#include <stdio.h>
#include <stdlib.h>

#define DEBUG	1
#define VERBOSE	1

#define PCM_BLOCK 		1
#define END_BLOCK		0
#define WAVHDR_SIZE		44
#define OUTPUT_FREQ		44100

#define BUFFER_REDLINE	0x100
#define BUFFER_SIZE	0x1000000


unsigned char *ad_data;
int  *ad_output;
char *ad_wavout;

char ad_wavhdr[WAVHDR_SIZE];

int  ad_sample;
int  ad_ref_index;
int  ad_ref_index_step_table[] = {
	-1, -1, -1, -1, 2, 4, 6, 8,
	-1, -1, -1, -1, 2, 4, 6, 8
};
int ad_ref_table[] = {
	0x010, 0x011, 0x013, 0x015, 0x017, 0x019, 0x01C, 0x01F,
	0x022, 0x025, 0x029, 0x02D, 0x032, 0x037, 0x03C, 0x042,
	0x049, 0x050, 0x058, 0x061, 0x06B, 0x076, 0x082, 0x08F,
	0x09D, 0x0AD, 0x0BE, 0x0D1, 0x0E6, 0x0FD, 0x117, 0x133,
	0x151, 0x173, 0x198, 0x1C1, 0x1EE, 0x220, 0x256, 0x292,
	0x2D4, 0x31C, 0x36C, 0x3C3, 0x424, 0x48E, 0x502, 0x583,
	0x610
};

int   decode(unsigned char code);
int   num_param(char *in, int *out);
int   implant_intval(char *out, int input, int num_bytes);
void  hexdump(char *in, int count);


int
main(int argc, char ** argv)
{
FILE *fin, *fout;

unsigned char data;
int ad_read;
int ad_write;
int i;

/* command-line parameters */
char infile_name[256];
char outfile_name[256];

int  samplerate_out  = OUTPUT_FREQ;
int  samplerate_in;
int  samplein_offset;
int  samplein_length;
int  amplifier;
int  num_samplein;
int  num_sampleout;

/* WAV file internals */
int  waveblock_size;
int  filesize_out;

/* WAV normalization & freq conversion */
int   max_val, min_val;
float multiplier, adjusted_val;
float time_value, slope;
float duration;
int   int_val;


#ifdef DEBUG
	/* print list of command-line parameters */
	for (i = 0; i < argc; i++)
	{
		printf("argv[%d] = %s\n", i, argv[i]);
	}
	printf("\n");
#endif

    if ((argc < 6) || (argc > 7))
	{
		printf("Invalid number of parameters : %d\n", argc);
		printf("Usage:\n");
		printf("adpcm <filein> <offset> <sample_len> <sample_freq> <fileout> [amplification]\n");
		exit(1);
	}

	strcpy(infile_name, argv[1]);
	strcpy(outfile_name, argv[5]);

	if (num_param(argv[2], &samplein_offset))
	{
		printf("Problem understanding offset %s\n", argv[2]);
		exit(1);
	}

	if (num_param(argv[3], &samplein_length))
	{
		printf("Problem understanding length %s\n", argv[3]);
		exit(1);
	}

	if (num_param(argv[4], &samplerate_in))
	{
		printf("Problem understanding samplerate %s\n", argv[4]);
		exit(1);
	}

	amplifier = 0;
	if (argc == 7)
	{
		if (num_param(argv[6], &amplifier))
		{
			printf("Problem understanding amplification %s\n", argv[6]);
			exit(1);
		}
	}


#ifdef DEBUG
	printf("Values:\n");
	printf("-------\n");
	printf("filein  = %s\n", infile_name);
	printf("offset  = 0x%X\n", samplein_offset);
	printf("length  = 0x%X\n", samplein_length);
	printf("freq    = %d\n", samplerate_in);
	printf("fileout = %s\n", outfile_name);
	if (argc == 7)
		printf("amplifier = %d\n", amplifier);
#endif

	num_samplein   = samplein_length * 2;	/* 2 per byte */
	duration       = ((float)num_samplein) / ((float)samplerate_in);
	num_sampleout  = (int)(duration * samplerate_out);

#ifdef VERBOSE
	printf("Duration of sample = %f seconds\n", duration );
#endif

	/* open input file */
	fin = fopen(infile_name, "rb");
	if (!fin)
	{
		printf("Couldn't open file %s\n", infile_name);
		exit(1);
	}

	/* open output file */
	fout = fopen(outfile_name, "wb");
	if (!fout)
	{
		printf("Couldn't open output file %s\n", outfile_name);
		exit(1);
	}

	/* Alloc buffers */
	ad_data   = (char *)malloc(num_samplein + BUFFER_REDLINE);
	ad_output = (int  *)malloc((num_samplein * (sizeof(int))) + BUFFER_REDLINE);
	ad_wavout = (char *)malloc((num_sampleout * 2 * 2) + BUFFER_REDLINE);


	/* read input file */
	fseek(fin, samplein_offset, SEEK_SET);
	fread(&ad_data[0], samplein_length, 1, fin);
	fclose(fin);

#ifdef DEBUG
	/*
	hexdump(&ad_data[0], 0x100);
	*/
#endif

	/* initialize data */
	ad_ref_index = 0;
	ad_sample = 0;
	ad_read = 0;
	ad_write = 0;

	/* translate ADPCM data */
	for(i=0;i<samplein_length;i++) {
		data = ad_data[ad_read++];
		ad_output[ad_write++] = decode((data >> 4) & 0x0F);
		ad_output[ad_write++] = decode((data) & 0x0F);
	}

#ifdef DEBUG
	/*
	hexdump((char *)(&ad_output[0]), 0x100);
	*/
#endif

	/* Normalize samples to output file */
	/* find max sample & multiplier for normalizing */
	max_val = 0;
	min_val = 0;

	for (i = 0; i < num_samplein; i++)
	{
		if (ad_output[i] < min_val)
			min_val = ad_output[i];

		if (ad_output[i] > max_val)
			max_val = ad_output[i];
	}

	if ((max_val - min_val) > 0)
	{
		multiplier = (0.75 * 65536) / (max_val - min_val);
	}

	if (argc == 7)
	{
		multiplier = amplifier;
	}

#ifdef VERBOSE
	printf("max_val = %d, min_val = %d\n", max_val, min_val);
	printf("max peak-to-peak = %d of possible 65536\n", (max_val - min_val));
	if (argc == 7)
		printf("adjusting by fixed-multiplier %f\n", multiplier);
	else
		printf("adjusting by multiplier %f\n", multiplier);
#endif

	/* do linear interpolation of samples for frequency conversion, and      */
	/* normalize to new sound level by mulitplying base sample by mulitplier */
	/* while copying to output buffer */

	for (i = 0; i < num_sampleout; i++)
	{
		int index;

		time_value = ((float)i) * ((float)samplerate_in) / ((float)samplerate_out);
		index = (int) time_value;
		time_value -= index;
		slope = time_value * (ad_output[index+1] - ad_output[index]);

		adjusted_val = multiplier * (ad_output[index] + slope);

		int_val = (int) adjusted_val;

		implant_intval(&ad_wavout[i*4], int_val, 2);
		implant_intval(&ad_wavout[i*4+2], int_val, 2);
	}

	/* create .WAV header */
	waveblock_size = num_sampleout * 4;
	filesize_out   = waveblock_size + WAVHDR_SIZE + 1;

	strcpy(&ad_wavhdr[0],  "RIFF");
	strcpy(&ad_wavhdr[8],  "WAVE");
	strcpy(&ad_wavhdr[12], "fmt ");
	strcpy(&ad_wavhdr[36], "data");

	implant_intval(&ad_wavhdr[4], filesize_out-8, 4);	/* filesize - 8 */
	implant_intval(&ad_wavhdr[16], 16, 4);		/* format chunk data length */
	implant_intval(&ad_wavhdr[20], PCM_BLOCK, 2);		/* sample data format */
	implant_intval(&ad_wavhdr[22], 2, 2);		/* # channels */
	implant_intval(&ad_wavhdr[24], samplerate_out, 4);	/* samplerate */
	implant_intval(&ad_wavhdr[28], samplerate_out * 4, 4);	/* avg # bytes/sec */
	implant_intval(&ad_wavhdr[32], 2, 2);	/* block alignment */
	implant_intval(&ad_wavhdr[34], 16, 2);	/* significant bits per sample */
	implant_intval(&ad_wavhdr[40], waveblock_size, 4);	/* len of sound data */


	/* write output file */
	fwrite(ad_wavhdr, WAVHDR_SIZE, 1, fout);
	fwrite(ad_wavout, waveblock_size, 1, fout);
	fputc(0, fout);

	fclose(fout);

	/* free buffers */
	free(ad_data);
	free(ad_output);
	free(ad_wavout);

	exit(0);
}

int
decode(unsigned char code)
{
	int ref;
	int val;

	ref = ad_ref_table[ad_ref_index];
	val = 0;

	if (code & 0x04)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x02)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x01)
		val += ref;
	ref  = (ref >> 1);
	val +=  ref;
	if (code & 0x08)
		val = -val;

	ad_sample += val;
	if (ad_sample < -2048)
		ad_sample = -2048;
	else
	if (ad_sample >  2047)
		ad_sample =  2047;

	ad_ref_index += ad_ref_index_step_table[code];
	if (ad_ref_index <  0)
		ad_ref_index =  0;
	else
	if (ad_ref_index > 48)
		ad_ref_index = 48;

	return (ad_sample);
}


int
implant_intval(char *out, int input, int num_bytes)
{
int i;
int temp = input;

	for (i = 0; i < num_bytes; i++)
	{
		*(out + i) = (char) (temp & 0xff);
		temp >>= 8;
	}
}


int
num_param(char *in, int *out)
{
int i, ret;
int radix, currval, tempint;
char temp;

	i = 0;
	ret = 0;

	radix = 10;
	currval = 0;

	if (strncmp(in, "0x", 2) == 0)
	{
		radix = 16;
		i = 2;
	}

	while ( *(in+i) != 0 )
	{
		temp = *(in+i);

		if ((temp >= '0') && (temp <= '9'))
		{
			tempint = (int) (temp - '0');
		}
		else if (radix == 16)
		{
			if ((temp >= 'A') && (temp <= 'F'))
			{
				tempint = (int) (temp - 'A') + 10;
			}
			else if ((temp >= 'a') && (temp <= 'f'))
			{
				tempint = (int) (temp - 'a') + 10;
			}
		}
		else
		{
			*out = currval;
			return(1);
		}

		currval *= radix;
		currval += tempint;
		i++;
	}
			
	*out = currval;
	return(0);
}

void
hexdump(char *in, int count)
{
int i, j;
unsigned char c;

	for (i = 0; i < count; i += 16)
	{
		printf("%4.4X: ", i);

		for (j = 0; j < 16; j++)
		{
			c = *(in+i+j);
			printf("%2.2X ", c);
		}

		printf("\n");
	}
}

