#include <stdio.h>
#include <stdlib.h>

#define DEBUG	1
#define VERBOSE	1

#define PCM_BLOCK 		1
#define END_BLOCK		0
#define WAVHDR_SIZE		44
/* #define OUTPUT_FREQ		44100 */

#define BUFFER_REDLINE	0x100
#define BUFFER_SIZE	0x1000000


unsigned char *ad_data;
int  *ad_input;
char *ad_wavin;

char ad_wavhdr[WAVHDR_SIZE];

int  ad_sample;
int  ad_ref_index;
int  ad_ref_index_step_table[] = {
	-1, -1, -1, -1, 2, 4, 6, 8,
	-1, -1, -1, -1, 2, 4, 6, 8
};
int ad_ref_table[] = {
	0x010, 0x011, 0x013, 0x015, 0x017, 0x019, 0x01C, 0x01F,
	0x022, 0x025, 0x029, 0x02D, 0x032, 0x037, 0x03C, 0x042,
	0x049, 0x050, 0x058, 0x061, 0x06B, 0x076, 0x082, 0x08F,
	0x09D, 0x0AD, 0x0BE, 0x0D1, 0x0E6, 0x0FD, 0x117, 0x133,
	0x151, 0x173, 0x198, 0x1C1, 0x1EE, 0x220, 0x256, 0x292,
	0x2D4, 0x31C, 0x36C, 0x3C3, 0x424, 0x48E, 0x502, 0x583,
	0x610
};

unsigned char   encode(int code);
int   decode(unsigned char code);
int   num_param(char *in, int *out);
int   grab_intval(char *in, int num_bytes);
int   implant_intval(char *out, int input, int num_bytes);
void  hexdump(char *in, int count);


int
main(int argc, char ** argv)
{
FILE *fin, *fout;

unsigned char data;
int ad_read;
int ad_write;
int i;

/* command-line parameters */
char infile_name[256];
char outfile_name[256];

/* int  samplerate_out = OUTPUT_FREQ; */
int  samplerate_out;
int  samplerate_in;
int  sampleout_offset;
int  sampleout_length;
float  amplifier;
int  num_samplein;
int  num_sampleout;

/* WAV file internals */
int  waveblock_size;
int  filesize_out;

/* WAV normalization & freq conversion */
int   max_val, min_val, peak_to_peak;
float multiplier, adjusted_val;
float time_value, slope;
float duration;
int   int_val;
int   temp, temp1, err;
int   temp_level;
int   byte;


#ifdef DEBUG
	/* print list of command-line parameters */
	for (i = 0; i < argc; i++)
	{
		printf("argv[%d] = %s\n", i, argv[i]);
	}
	printf("\n");
#endif

    if ((argc < 6) || (argc > 7))
	{
		printf("Invalid number of parameters : %d\n", argc);
		printf("Usage:\n");
		printf("adpcm <filein> <fileout> <offset> <sample_len> <samplerate_out> [amplification]\n");
		exit(1);
	}

	strcpy(infile_name, argv[1]);
	strcpy(outfile_name, argv[2]);

	if (num_param(argv[3], &sampleout_offset))
	{
		printf("Problem understanding offset %s\n", argv[3]);
		exit(1);
	}

	if (num_param(argv[4], &sampleout_length))
	{
		printf("Problem understanding length %s\n", argv[4]);
		exit(1);
	}

	if (num_param(argv[5], &samplerate_out))
	{
		printf("Problem understanding samplerate_out %s\n", argv[5]);
		exit(1);
	}

	amplifier = 0;
	if (argc == 7)
	{
		amplifier = strtod(argv[6], NULL);
		if (amplifier == 0)
		{
			printf("Problem understanding amplification %s\n", argv[6]);
			exit(1);
		}
	}

	/* get samplerate_in from WAV file itself */

#ifdef DEBUG
	printf("Values:\n");
	printf("-------\n");
	printf("filein  = %s\n", infile_name);
	printf("offset  = 0x%X\n", sampleout_offset);
	printf("length  = 0x%X\n", sampleout_length);
	printf("freq_out= %d\n", samplerate_out);
	printf("fileout = %s\n", outfile_name);
	if (argc == 7)
		printf("amplifier = %f\n", amplifier);
#endif


	/* open input file */
	fin = fopen(infile_name, "rb");
	if (!fin)
	{
		printf("Couldn't open file %s\n", infile_name);
		exit(1);
	}

	/* open output file */
	fout = fopen(outfile_name, "r+b");
	if (!fout)
	{
		fout = fopen(outfile_name, "wb");
		if (!fout)
		{
			printf("Couldn't open output file %s\n", outfile_name);
			exit(1);
		}
	}

	if (fseek(fout, sampleout_offset, SEEK_SET) != 0)
	{
		printf("Error in file offset - cannot locate in file\n");
		exit(1);
	}
	

	/* Check WAV files Parameters */

	for (i = 0; i < WAVHDR_SIZE; i++)
	{
		byte = fgetc(fin);

		if (byte == EOF)
		{
			printf("Could not read WAV header.  Bytes desired = %d; bytes read = %d\n", WAVHDR_SIZE, i);
			exit(1);
		}
		else
		{
			ad_wavhdr[i] = byte;
		}
	}

	if (memcmp(&ad_wavhdr[0],  "RIFF", 4) != 0)
	{
		printf("RIFF not found\n");
		exit(1);
	}
	if (memcmp(&ad_wavhdr[8],  "WAVE", 4) != 0)
	{
		printf("RIFF not found\n");
		exit(1);
	}
	if (memcmp(&ad_wavhdr[12], "fmt ", 4) != 0)
	{
		printf("RIFF not found\n");
		exit(1);
	}
	if (memcmp(&ad_wavhdr[36], "data", 4) != 0)
	{
		printf("RIFF not found\n");
		exit(1);
	}

	temp =  grab_intval(&ad_wavhdr[16], 4);		/* format chunk data length */
	if (temp != 16)
	{
		printf("format chunk data chunk length wrong: found 0x%X, expected 0x%X\n", temp, 16);
		exit(1);
	}

	if (grab_intval(&ad_wavhdr[20], 2) != PCM_BLOCK)		/* sample data format */
	{
		printf("sample data format wrong\n");
		exit(1);
	}

	/*
	temp = grab_intval(&ad_wavhdr[32], 2);	* block alignment *
	if (temp != 2)
	{
		printf("block alignment wrong: found 0x%X, expected 0x%x\n", temp, 2);
		exit(1);
	}
	*/

	if (grab_intval(&ad_wavhdr[34], 2) != 16)	/* significant bits per sample */
	{
		printf("significant bits per sample wrong\n");
		exit(1);
	}

	if (grab_intval(&ad_wavhdr[22],  2) != 2)		/* # channels */
	{
		printf("Number of channels wrong\n");
		exit(1);
	}

	filesize_out = grab_intval(&ad_wavhdr[4], 4) + 8;	/* filesize - 8 */

	samplerate_in = grab_intval(&ad_wavhdr[24], 4);	/* samplerate */
	printf("samplerate_in = %d\n", samplerate_in);

	/*
	temp = grab_intval(&ad_wavhdr[28], 4);	* avg # bytes/sec *
	if (temp != (samplerate_in * 4))
	{
		printf("average bytes per second wrong.  Found = 0x%X, expected 0x%X\n", temp, (samplerate_in * 4));
		exit(1);
	}
	*/

	waveblock_size = grab_intval(&ad_wavhdr[40], 4);	/* len of sound data */

	if ( sampleout_length == 0 ) {
		sampleout_length = waveblock_size / 8 ;
	}

	num_sampleout  = sampleout_length * 2;	/* 2 per byte */
	duration       = ((float)num_sampleout) / ((float)samplerate_out);

#ifdef VERBOSE
	printf("Duration of sample = %f seconds\n", duration );
#endif

	num_samplein   = (int)(duration * samplerate_in);
/*	waveblock_size = num_samplein * 4; */
/*	filesize_out   = waveblock_size + WAVHDR_SIZE + 1; */

	/* Alloc buffers */
	ad_data   = (char *)malloc(num_sampleout + BUFFER_REDLINE);
	ad_input  = (int  *)malloc((num_samplein * (sizeof(int))) + BUFFER_REDLINE);
	ad_wavin  = (char *)malloc((waveblock_size) + BUFFER_REDLINE);

	printf("waveblock_size = %d\n", waveblock_size);
	printf("num_samplein = %d\n", num_samplein);

	/* read input file */
	for (i = 0; i < waveblock_size; i++)
	{
		byte = fgetc(fin);

		if (byte == EOF)
		{
			printf("Could not read WAV input file.  Bytes desired = %d; bytes read = %d\n", waveblock_size, i);
			exit(1);
		}
		else
		{
			ad_wavin[i] = byte;
		}
	}

	fclose(fin);


	/* Normalize samples to output file */
	/* find max sample & multiplier for normalizing */
	max_val = 0;
	min_val = 65535;

	for (i = 0; i < num_samplein; i++)
	{
		temp_level = grab_intval(&ad_wavin[i*4], 2) + grab_intval(&ad_wavin[i*4+2], 2);

		if (temp_level < min_val)
			min_val = temp_level;

		if (temp_level > max_val)
			max_val = temp_level;
	}

	if ((max_val - min_val) > 0)
	{
		peak_to_peak = (min_val > 0) ? min_val : ((-1)*(min_val));
		peak_to_peak = (max_val > peak_to_peak) ? max_val : peak_to_peak;
		/* peak_to_peak *= 2; */

		multiplier = (float)(peak_to_peak) / 4096;
	}

	if (argc == 7)
	{
		multiplier = amplifier;
	}

#ifdef VERBOSE
	printf("max_val = %d, min_val = %d\n", max_val, min_val);
	printf("max peak-to-peak = %d of possible 65536\n", peak_to_peak);
	if (argc == 7)
		printf("adjusting by fixed-multiplier %f\n", multiplier);
	else
		printf("adjusting by multiplier %f\n", multiplier);
#endif

	/* do linear interpolation of samples for frequency conversion, and      */
	/* normalize to new sound level by mulitplying base sample by mulitplier */
	/* while copying to output buffer */

	for (i = 0; i < num_sampleout; i++)
	{
		int index;

		time_value = ((float)i) * ((float)samplerate_in) / ((float)samplerate_out);
		index = (int) time_value;
		time_value -= index;

		temp   = grab_intval(&ad_wavin[index*4], 2);
		temp  += grab_intval(&ad_wavin[index*4+2], 2);

		temp1  = grab_intval(&ad_wavin[(index+1)*4], 2);
		temp1 += grab_intval(&ad_wavin[(index+1)*4+2], 2);

		slope = time_value * (temp1 - temp);

		/* multiplier is actually a divider */
		adjusted_val = (temp + slope) / (2 * multiplier);

		int_val = (int) adjusted_val;

		ad_input[i] = int_val;
	}

#ifdef DEBUG
	/*
	hexdump(&ad_data[0], 0x100);
	*/
#endif

	/* initialize data */
	ad_ref_index = 0;
	ad_sample = 0;
	ad_read = 0;
	ad_write = 0;

	/* translate ADPCM data */
	for(i = 0; i < sampleout_length; i++)
	{
		data  = (encode(ad_input[ad_read++]) & 0x0F) << 4;
		data |= (encode(ad_input[ad_read++]) & 0x0F);
		ad_data[ad_write++] = data;
	}

#ifdef DEBUG
	/*
	hexdump((char *)(&ad_input[0]), 0x100);
	*/
#endif

	/* write output file */
	err = fseek(fout, sampleout_offset, SEEK_SET);
/*	if (err < 0)  */
	err = fwrite(ad_data,  sampleout_length, 1, fout);

	fclose(fout);

	/* free buffers */
	free(ad_data);
	free(ad_input);
	free(ad_wavin);

	exit(0);

}

unsigned char
encode(int wav_sample)
{
	int i;
	int mingap, mingapidx;
	int tst;
	int diff;
	int gapdiff;

	int ref;
	int val;
	int code;

	ref  = ad_ref_table[ad_ref_index];
	val  = 0;
	code = 0;


	/* check delta between last synthetic sample and next incoming sample     */
	/* bit 3 of val, if set, means the delta amount is negative, not positive */

	diff = wav_sample - ad_sample;
	if (diff < 0)
	{
		code |= 0x08;
		diff = -diff;
	}


	/* some default values to ensure that a real minimum delta */
	/* is achieved through comparison */

	mingap = 32767;
	mingapidx = 0;


	/* find all possible values based on predetermined table value */
	/* and use the most appropriate one.  Note that we constantly  */
	/* aim to approximate the original waveform, not just to       */
	/* approximate deltas */

	for (i = 0; i < 8; i++)
	{
		tst = ref>>3;
		if (i & 0x01)
			tst += (ref>>2);
		if (i & 0x02)
			tst += (ref>>1);
		if (i & 0x04)
			tst += ref;

		gapdiff = diff - tst;
		if (gapdiff < 0)
			gapdiff = -gapdiff;

		if (gapdiff < mingap)
		{
			mingap = gapdiff;
			mingapidx = i;
		}
	}

	code |= mingapidx;

	if (code & 0x04)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x02)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x01)
		val += ref;
	ref  = (ref >> 1);
	val +=  ref;
	if (code & 0x08)
		val = -val;

	ad_sample += val;
	if (ad_sample < -2048)
		ad_sample = -2048;
	else
	if (ad_sample >  2047)
		ad_sample =  2047;

	ad_ref_index += ad_ref_index_step_table[code];
	if (ad_ref_index <  0)
		ad_ref_index =  0;
	else
	if (ad_ref_index > 48)
		ad_ref_index = 48;

	return (code);
}


int
decode(unsigned char code)
{
	int ref;
	int val;

	ref = ad_ref_table[ad_ref_index];
	val = 0;

	if (code & 0x04)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x02)
		val += ref;
	ref = (ref >> 1);
	if (code & 0x01)
		val += ref;
	ref  = (ref >> 1);
	val +=  ref;
	if (code & 0x08)
		val = -val;

	ad_sample += val;
	if (ad_sample < -2048)
		ad_sample = -2048;
	else
	if (ad_sample >  2047)
		ad_sample =  2047;

	ad_ref_index += ad_ref_index_step_table[code];
	if (ad_ref_index <  0)
		ad_ref_index =  0;
	else
	if (ad_ref_index > 48)
		ad_ref_index = 48;

	return (ad_sample);
}


int
grab_intval(char *in, int num_bytes)
{
int i;
int temp = 0;
int multiplier = 1;
unsigned char byte;


	for (i = 0; i < num_bytes; i++)
	{
		byte = *(in + i);
		temp += byte * multiplier;
		multiplier <<= 8;
	}

	/* sign-extend a negative 2-byte number */
	if ((num_bytes == 2) && (temp & 0x8000))
	{
		temp |= 0xFFFF0000;
	}
	return(temp);
}


int
implant_intval(char *out, int input, int num_bytes)
{
int i;
int temp = input;

	for (i = 0; i < num_bytes; i++)
	{
		*(out + i) = (char) (temp & 0xff);
		temp >>= 8;
	}
}


int
num_param(char *in, int *out)
{
int i, ret;
int radix, currval, tempint;
char temp;

	i = 0;
	ret = 0;

	radix = 10;
	currval = 0;

	if (strncmp(in, "0x", 2) == 0)
	{
		radix = 16;
		i = 2;
	}

	while ( *(in+i) != 0 )
	{
		temp = *(in+i);

		if ((temp >= '0') && (temp <= '9'))
		{
			tempint = (int) (temp - '0');
		}
		else if (radix == 16)
		{
			if ((temp >= 'A') && (temp <= 'F'))
			{
				tempint = (int) (temp - 'A') + 10;
			}
			else if ((temp >= 'a') && (temp <= 'f'))
			{
				tempint = (int) (temp - 'a') + 10;
			}
		}
		else
		{
			*out = currval;
			return(1);
		}

		currval *= radix;
		currval += tempint;
		i++;
	}
			
	*out = currval;
	return(0);
}

void
hexdump(char *in, int count)
{
int i, j;
unsigned char c;

	for (i = 0; i < count; i += 16)
	{
		printf("%4.4X: ", i);

		for (j = 0; j < 16; j++)
		{
			c = *(in+i+j);
			printf("%2.2X ", c);
		}

		printf("\n");
	}
}

